/*
MP3 STATION

A plain text Linux/Unix toolset for managing mp3 playback with as little
user interaction as possible.  It was built to handle mp3 playback in a car.
Copyright (C) 2003 David Rao <louigi600@yahoo.it>

This file is part of MP3 STATION.

MP3 STATION is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MP3 STATION is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MP3 STATION; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include <term.h>
#include <stdlib.h>

#ifdef LCURS
#include LCURS
#else
#include <ncurses.h>
#endif

#include "my_functions.h"

#define MAX_PATH 1024

typedef struct directory_listing {
  char *alloc_ptr;
  long n_rows;
  int  r_size;
} dir_list;


int VERBOSE=1;        //verbose level
#ifdef TMPD
char TMP_DIR[MAX_PATH]=TMPD;
#else
char TMP_DIR[MAX_PATH]="/tmp/cm3s";
#endif
char STAT_PIPE[MAX_PATH]="stat";   //this must be prefixed with TMP_DIR
extern char PROJ_NAME;

WINDOW *w,*w2,*w3,*w4; //windows 
int split;             //no of lines in witch main window is split
int cols,lins;         //line and column size of main window
long count=0,count1=0; //counter for number of entries found in a directry
int max_len=0;         //actual max lenght found while counting
char trk[MAX_PATH],buf[MAX_PATH];
int stat_fd[2];  
dir_list dirlist={NULL,0,0};
help playback_help={
  "e,esc: Exit client       , R   : Toggle repeat , arrorL : Restart track",
  "p    : Toggle pause      , l   : Load menu     , arrowR : Skip track",
  "q    : Quit cm3s         , pgU : Back pg tracks, arrowU : Prev. track",
  "r    : Toggle random     , pgD : Skip pg tracks, arrowD : Skip track",
  "R    : Toggle repeat     , b   : Toggle reverse, m      : Mixer",
  ""
};
help load_help={
  "e,esc: : Exit load menu and return to main menu",
  "arrorL : Move up a directory , arrowR : Descend into selected directory",
  "arrowU : Scroll up           , arrowD : Scroll down",
  "e      : Exit load menu and return to main menu.",
  "ENTER  : Load selected file/playlist/directory ",
  ""
};

//***********************************************************
int Is_Playable (char str[MAX_STR]);
void Load ();
long Count_Calls (char str[MAX_STR]); //count calls for alloc mem (dirwalk)  
long Store_Ent (char str[MAX_STR]);  //after count store in alloc mem (dirwalk)
int Wshow_Mem (WINDOW *win, dir_list content);
int wshow_help (WINDOW *win,help hlp_matrix);

//***********************************************************
int wshow_help (win,hlp_matrix)
WINDOW *win;
help hlp_matrix;
{ int i=0;
 
  clearok(win,TRUE);
  wclear(win);
  wrefresh(win);
  clearok(win,FALSE); 
  scrollok(win,FALSE);
  while(strcmp(hlp_matrix[i],"")!=0)
  { wprintw(win,"%s\n",hlp_matrix[i]);
    ++i;
  }
  wrefresh(win);
  scrollok(win,TRUE);
}

//***********************************************************
void Load ()
{ char *ptr;
  char tbuf2[MAX_PATH];

  readmsg(PROJ_NAME,tbuf2); 
//  while(readmsg(PROJ_NAME,tbuf2)!=-1) { sleep(1); }
  while(readmsg(PROJ_NAME,tbuf2)!=-1) { }
  sendmsg(PROJ_NAME,trk);
  wprintw(w3,"l: Loading playlist!\n");
  getcwd(buf,MAX_PATH);
  count=0;
  max_len=0;
  dirwalk(buf,Count_Calls);
  wprintw(w3,"%i Playable files found, max lenght %i\n",count,max_len);
  dirlist.alloc_ptr=NULL;
  if((ptr=(char*)calloc(count,max_len))!=NULL)
  { dirlist.alloc_ptr=ptr;
    dirlist.n_rows=count;
    dirlist.r_size=max_len;
    wprintw(w3,"Allocated %3.2f Kb for directory listing!\n",
      (float)(count * max_len)/1024);
    wrefresh(w3);
    count1=0;
    dirwalk(buf,Store_Ent);
    Wshow_Mem(w2,dirlist);
  } else wprintw(w3,"Failed to allocate memory!\n");
//  if(dirlist.alloc_ptr!=NULL) free(dirlist.alloc_ptr);
}

//***********************************************************
long Count_Calls (str)
char str[MAX_STR];
{ long t;
  if(Is_Playable(str)==0) 
  { ++count;
    if((t=(strlen(str)+1)) > max_len) max_len=t;  //one extra char for \0
  } 
  return(0); 
}

//***********************************************************
long Store_Ent (str)
char str[MAX_STR];
{ char *ptr2;
  if(Is_Playable(str)==0)
  { ptr2=dirlist.alloc_ptr + (dirlist.r_size * count1);
     sprintf((char*)ptr2,"%s\0",str);
    ++count1;
  }
}

//***********************************************************
int Is_Playable (name)
char name[MAX_STR];
{   char *sp;
  char tmps[MAX_STR];

  if (Is_File(name)==0)
  { if((sp=strrchr(name,'.'))!=NULL)
    { strtolower((char*)++sp,tmps);
      if((strcmp(tmps,"mp3")==0) || (strcmp(tmps,"msl")==0))
        return(0);
      else return(-1);
    } else return(-1);
  } 
  else if(Is_Dir(name)==0) return(0);  
  else return(-1); 
}

//***********************************************************
int Wshow_Mem (win,content)
WINDOW *win;
dir_list content;
{ long i=0,j=0;
  char *ptr2=NULL;
  int x,y,c;
  int h_line=0;      //selected line for highliting
  int first_line=0;  //first line displayed in window

  getmaxyx(win,y,x);
//  wprintw(win,"%i,%i\n",y,x);

  nodelay(w3, FALSE);
  scrollok(win,FALSE);
  while (1) 
  { clearok(win,TRUE);
    wclear(win);
    wrefresh(win);
    clearok(win,FALSE);
    for(i=0;i<min((content.n_rows - first_line),y);i++)
    { j=i + first_line; 
      ptr2=content.alloc_ptr + (content.r_size * j);
      if(i==h_line) wattron(win,A_STANDOUT);
      wprintw(win,"%s\n",(char*)ptr2);
      if(i==h_line) wattroff(win,A_STANDOUT);
    }
    wrefresh(win);
    wrefresh(w3);
    c=wgetch(w3);
//Arrow left
    if (c==KEY_LEFT)
    { wprintw(w3,"<-: Parent dir listing!\n");
      if( content.alloc_ptr != NULL ) free(content.alloc_ptr);
      chdir("..");
      Load();
      break;
    }
//Arrow right
    else if (c==KEY_RIGHT)
    { wprintw(w3,"->: Descend into directory!\n");
      ptr2=content.alloc_ptr + (content.r_size * (first_line + h_line));
      if(Is_Dir((char*)ptr2)==0)
      { chdir((char*)ptr2);
        if( content.alloc_ptr != NULL ) free(content.alloc_ptr);
        Load();
        break;
      }
    }
//Arrow up
    else if (c==KEY_UP)
    { waddch(w3,ACS_UARROW);
      wprintw(w3,": Scroll up\n",ACS_UARROW);
      if(h_line>0) --h_line;
      else if(first_line>0) --first_line; 
    }
//Arrow down
    else if (c==KEY_DOWN)
    { waddch(w3,ACS_DARROW);
      wprintw(w3,": Scroll down!\n");
      if(h_line < min((y - 1),(content.n_rows - first_line - 1))) ++h_line;
      else if(first_line<(content.n_rows - 1)) 
      { ++first_line;
        if(h_line> (content.n_rows - first_line - 1)) --h_line;  
      }
    }
//Enter
    if (c==13)
    { wprintw(w3,"ENTER: Plying selection!\n");
      ptr2=content.alloc_ptr + (content.r_size * (first_line + h_line));
      sprintf(buf,"NOP\n#L %s\n",(char*)ptr2);
      write(stat_fd[1],buf,strlen((char*)ptr2)+8);
      if( content.alloc_ptr != NULL ) free(content.alloc_ptr);
      break;
    }
//101 = e for exiting load menu
    else if (c==101)
    { wprintw(w3,"%c: Exiting load menu!\n",c);
      wrefresh(w3);
      sleep(1);
      if( content.alloc_ptr != NULL ) free(content.alloc_ptr);
      break;
    }
//104 = h for help on present menu
  else if (c==104) wshow_help(w3,load_help);
  }
  nodelay(w3, TRUE);
  clearok(win,TRUE);
  wclear(win);
  wrefresh(win);
  clearok(win,FALSE);
  scrollok(win,TRUE);
}

//***********************************************************
//initializing window stuff for curses to work
int Init_Windows (percent)
int percent;
{ char buf2;

  w=initscr();            /* initializing window */
  getmaxyx(w,lins,cols); 
//this makes sure that the bottomb window can display at leat 5 lines
//note that 2 extra are necesarry for boxing the window (5+2=7)
//  split=max((floor((lins * percent) / 100)),(lins  - 7));
  split=max(((lins * percent) / 100),(lins  - 7));
  keypad(w, FALSE);
  cbreak();               /* no line buffering */
//  nodelay(w, TRUE);     /* make getch call non blocking */
  noecho();               /* no echoing keys pressed */
  nonl();                 /* no new line translation in windows */
  box(w,ACS_VLINE,ACS_HLINE);
  move(split,1);
  hline(ACS_HLINE,cols-2);
  move(split,0);
  addch(ACS_LTEE);
  move(split,cols -1);
  addch(ACS_RTEE);
  wrefresh(w);
  w2=newwin((split - 1),(cols - 2),1,1);
  scrollok(w2,TRUE);
//  box(w2,ACS_VLINE,ACS_HLINE);
  w3=newwin((lins - split - 2),(cols - 2),(split + 1),1);
  scrollok(w3,TRUE);
//  box(w3,ACS_VLINE,ACS_HLINE);
  wmove(w2,0,0);
//  wprintw(w2,"%i,%i,%i,%i\n",split,lins,cols,sizeof(&w));
  keypad(w2, FALSE);
  wrefresh(w2);
  wmove(w3,0,0);
  wrefresh(w3);
  nodelay(w3, TRUE);
  keypad(w3,TRUE);
//  keypad(w2,FALSE);
}

//***********************************************************
//manage the keypad during track management mode
int Trk_Keypad_Mgmt ()
{ int c,d;
  int x,y;
  int ti;
  char tbuf[MAX_PATH];

  wrefresh(w2);
  wrefresh(w3);
  c=wgetch(w3);
  nodelay(w3, TRUE);
  if (c==ERR) { sleep(1); return(0); }
//these are the keys I managed to get working quickly
//Arrow left
  else if (c==KEY_LEFT)
  { wprintw(w3,"\n<-: Restarting track!");
    write(stat_fd[1],"NOP\n#E RESTART\n",15);
  }
//Arrow right
  else if (c==KEY_RIGHT)
  { wprintw(w3,"\n->: Sckipping to next track!");
    write(stat_fd[1],"NOP\n@P 0\n",9);
  }
//Arrow up
  else if (c==KEY_UP)
  { waddch(w3,10); 
    waddch(w3,ACS_UARROW);
    wprintw(w3,": Sckipping to previous track!");
    write(stat_fd[1],"NOP\n#E PREV\n",12);
  }
//Arrow down
  else if (c==KEY_DOWN)
  { waddch(w3,10);
    waddch(w3,ACS_DARROW);
    wprintw(w3,": Sckipping to next track!");
    write(stat_fd[1],"NOP\n@P 0\n",9);
  }
//Page up
  else if (c==KEY_PPAGE)
  { getmaxyx(w2,y,x);
    wprintw(w3,"\nPG UP: Skipping back %i tracks!",y);
    sprintf(tbuf,"NOP\n#E SKIP -%i\n",y);
    write(stat_fd[1],&tbuf,17);
  }
//Page down
  else if (c==KEY_NPAGE)
  { getmaxyx(w2,y,x);
    wprintw(w3,"\nPG DOWN: Skipping forward %i tracks!",y);
    sprintf(tbuf,"NOP\n#E SKIP %i\n",y);
    write(stat_fd[1],&tbuf,16);
  }
//Function keys F5 throught to F12
  else if (c==KEY_F(5))
  { wprintw(w3,"\nF5: Not yet mapped to enything!");
  }
  else if (c==KEY_F(12))
  { wprintw(w3,"\nF12: Not yet mapped to enything!");
  }
//Normal keys can be trapped with ascii code
//82 = R togle repeat playlyst playback mode
  else if (c==82)
  { wprintw(w3,"\n%c: Toggling repeat function status ; ",c);
    write(stat_fd[1],"NOP\n#E REPEAT\n",14);
    sleep(1);
    if(readmsg(PROJ_NAME,tbuf)!=-1) wprintw(w3,"%s",&tbuf);
  }
//98 = b reverse playback order
  else if (c==98)
  { wprintw(w3,"\n%c: Toggling order ; ",c);
    write(stat_fd[1],"NOP\n#E BACKPLAY\n",17);
//    sleep(1);
//    if(readmsg(PROJ_NAME,tbuf)!=-1) wprintw(w3,"%s",&tbuf);
    while(readmsg(PROJ_NAME,tbuf)==-1) { sleep(1); }
    wprintw(w3,"%s",&tbuf);
  }
//101 = e for exiting client without quitting cm3s
  else if (c==101)
  { wprintw(w3,"\n%c: Exiting without terminating cm3s!",c);
    wrefresh(w3);
    sleep(1);
    return(-1);
  }
//104 = h for help on present menu
  else if (c==104) wshow_help(w3,playback_help);
//105 = i for getting id3tags
  else if (c==105) 
  { write(stat_fd[1],"NOP\n#E SHOWID3TAG\n",18);
    while(readmsg(PROJ_NAME,tbuf)==-1) { sleep(1); }
    clearok(w3,TRUE);
    wclear(w3);
    wrefresh(w3);
    clearok(w3,FALSE);
    scrollok(w3,FALSE);
    wprintw(w3,"%s",&tbuf);
    wrefresh(w3);
    scrollok(w3,TRUE);
  }
//108 = l for loading new file/playlist
  else if (c==108) Load();
#ifdef MIXER
//109 = m for mixer
  else if (c==109) 
  { while(readmsg(PROJ_NAME,tbuf)!=-1) { sleep(1); }
    mixer();
    sendmsg(PROJ_NAME,trk); 
  }
#endif
//112 = p for toggling pause
  else if (c==112)
  { wprintw(w3,"\n%c: Toggling pause!",c);
    write(stat_fd[1],"NOP\n#E PAUSE\n",13); 
  }
//113 = q for quitting
  else if (c==113)
  { wprintw(w3,"\nq: Quittint and terminating cm3s!");
    wrefresh(w3);
    sleep(1);
    write(stat_fd[1],"NOP\n#E QUIT\n",12);
    return(-1);
  }
//114 = r for random play
  else if (c==114)
  { wprintw(w3,"\n%c: Randomizing list ... ",c);
    write(stat_fd[1],"NOP\n#E RAND\n",12);
//    sleep(1);
//    if(readmsg(PROJ_NAME,tbuf)!=-1) wprintw(w3,"%s",&tbuf);
    while(readmsg(PROJ_NAME,tbuf)==-1) { sleep(1); }
    wprintw(w3,"%s",&tbuf);
   }
//115 = s for message sending test
else if (c==115)
  { wprintw(w3,"\n%c: Sending a test message via ipc!",c);
    sendmsg(PROJ_NAME,"test");
  }
//116 = t for message reading test
else if (c==116)
  { wprintw(w3,"\n%c: Reding any message left in ipc message queue ... ",c);
    readmsg(PROJ_NAME,buf);
    wprintw(w2,"%s",&buf);
  }
//This traps artificially ESC and correctly pusces pack in place any
//unhandled escape sequence
//esc also quits
  else if (c==27)
  { c=wgetch(w3);
    if(c==ERR)
    { wprintw(w3,"\nESC: Exiting without terminating cm3s!");
      wrefresh(w3);
      sleep(1);
      return(-1);
    }
  }
  else wprintw(w3,"\n%c(%i): ;-P",c,c);
  wrefresh(w3);
  wrefresh(w2);
//  printf("%c",7);
}

//***********************************************************
int main ()
{ char ts[MAX_PATH];
  int c,d;
  int x,y;
  int tmp;

  drop_privs();
  sprintf(ts,"%s/%s",TMP_DIR,STAT_PIPE);
  strcpy(STAT_PIPE,ts);

//Opening status pipe in normal mode
//this will hang until sopething opens the pipe for reading
//if cm3s has already been started no problem
//if not no problem it will hang untill you start it
stat_fd[1]=open(STAT_PIPE,O_WRONLY);

Init_Windows(71);

//This displays some basic information on the GPL in the bottomb window
wprintw(w3,"MP3 STATION project, Copyright (C) 2003 David Rao\n");
wprintw(w3,"MP3 STATION comes with ABSOLUTELY NO WARRANTY.\n");
wprintw(w3,"This is free software, and you are welcome to redistribute it\n");
wprintw(w3,"under the conditions of the GNU General Public License; read COPYING.\n");

while(1)
{
  if(readmsg(PROJ_NAME,buf)!=-1) 
  { strcpy(trk,buf); 
    wprintw(w2,"%s\n",&trk);
  }
  if(Trk_Keypad_Mgmt()==-1) break;
}
endwin();
close(stat_fd[1]);
sendmsg(PROJ_NAME,trk);
return(0);
}
