/*
MP3 STATION

A plain text Linux/Unix toolset for managing mp3 playback with as little
user interaction as possible.  It was built to handle mp3 playback in a car.
Copyright (C) 2003 David Rao <louigi600@yahoo.it>

This file is part of MP3 STATION.

MP3 STATION is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MP3 STATION is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MP3 STATION; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <signal.h>
#include <string.h>

//this is only here to be able to include the mixer stuff
#ifdef MIXER
#ifdef LCURS
#include LCURS
#else
#include <ncurses.h>
#endif
#endif

#include "my_functions.h"


#define MAX_PATH 1024

typedef struct playlist_handle {
  char *alloc_ptr;
  long num_rows;
  int row_size;
} ply_hdl;


#ifdef MP3D
  char MP3_DIR[MAX_PATH]=MP3D;
#else
  char MP3_DIR[MAX_PATH]="/mnt/usb_hd/mp3";
#endif

#ifdef TMPD
  char TMP_DIR[MAX_PATH]=TMPD;
#else
  char TMP_DIR[MAX_PATH]="/tmp/cm3s";
#endif

extern char PROJ_NAME;
char pt[10]="server";
//This will be in the user that runs the command home and not cm3s user home
char INI_FILE[MAX_PATH]="cm3s.ini"; 
char STAT_PIPE[MAX_PATH]="stat";     //this must be prefixed with TMP_DIR
char CMD_PIPE[MAX_PATH]="cmd";       //this must be prefixed with TMP_DIR
char SETTINGS[MAX_PATH]="config";    //this must be prefixed with TMP_DIR
char PLAYLIST[MAX_PATH]="last.msl";  //this must be prefixed with TMP_DIR
char LOG[MAX_PATH]="log";            //this must be prefixed with TMP_DIR
char SAVELIST[MAX_PATH]="last.msl";  //this must be prefixed with TMP_DIR
char TPL[MAX_PATH]="temp.msl";       //this must be prefixed with TMP_DIR
char OFFSET[MAX_PATH]="last.trk";    //this must be prefixed with TMP_DIR
//Remember that the first thing that must be done in main after  
//parsing options is to prefix the correct paths to the variables
//that need the directory prefix to be a correct absolute path
#ifdef VER
  char VERSION[80]=VER;
#else
  char VERSION[80]="Version 1.0 beta";
#endif
char MASTER_VOL[]="67,67  ";   //default stereo master volume % 
char PCM_VOL[]="80,80  ";      //default stereo pcm volume % 
int BEDAEMON=0;
int BUFFER=0;
int RANDOM=0;
int REPEAT=1,repeat;
int BACKPLAY=0,backplay;
long START=-1; //this really is 0 but -1 will distinguish default from request
long SFRAME=0;  //1 = concider the saved frame when you start
int NEXT=0;    //1 = skip a track at start implies SFRAME=0
int LPT_IO=0;
int VERBOSE=1; //for now the default verbose level is 1

int a1,a2;
int pid,pid1,pid2,spc_pid,ec_pid=0;
int cmd_fd[2],stat_fd[2];
int status;
ply_hdl playlist;
char tbuf[MAX_PATH];
char hist_name[MAX_PATH];
key_t hist_list;
int ugids[2];
long frame,cur_track;
char progname[255];

//this is only here to be able to include the mixer stuff
#ifdef MIXER
WINDOW *w,*w2,*w3,*w4; //windows
int wshow_help (win,hlp_matrix)
{ return(0); }
#endif

//*******************************************************
void sig_quit_handler ();
void check();
void save_info();
int Start_Server ();
int Stop_Server ();
void Read_Dir (char name[MAX_PATH] , int fd);
ply_hdl Load_Playlist (char name[MAX_PATH], int rand);
ply_hdl Randomize_Playlist (ply_hdl playlist, int rand);
int Save_Playlist (ply_hdl playlist, char where[MAX_PATH], int randseed);
int Load_File (char name[MAX_PATH]);
int Manage_File (long offset, ply_hdl playlist, long trackno);
int Manager (ply_hdl playlist, long start_track);
int main(int argc, char **argv);

//*******************************************************
void sig_quit_handler ()
{ Stop_Server();
  exit(0);
}

//*******************************************************
void check()
{ char pn[256];
  int pid=0;

#ifdef MPG
  strcpy(pn,(char*)basename(MPG));
#else
  strcpy(pn,"mpg123");
#endif

//    if((other_running(pn)!=0) || (other_running(progname)!=0)) 
    if((other_running(pn)!=0) || (other_running(progname)!=0) || (fuser("/dev/dsp")!=0))
    { printf("Aborting: already running or /dev/dsp is in use!\n");
      exit(1);
    }
}

//*******************************************************
//this saves the current track and frame
void save_info()
{ int tfd;
  char tmps[MAX_PATH];

  tfd=open(OFFSET,O_WRONLY|O_CREAT|O_TRUNC,0666);
  write(tfd,&cur_track,sizeof(cur_track));
  write(tfd,&frame,sizeof(frame));
  close(tfd);
  sprintf(tmps,"%s/%s.trk",TMP_DIR,hist_name);
  tfd=open(tmps,O_WRONLY|O_CREAT|O_TRUNC,0666);
  write(tfd,&cur_track,sizeof(cur_track)); 
  write(tfd,&frame,sizeof(frame));
  close(tfd);
}

//*******************************************************
ply_hdl Randomize_Playlist (playlist,rand)
ply_hdl playlist;
int rand;
{ int max=0,l=0,tpfd[2];
  long count=0,i=0;
  char *ptr,*ptr2;
  char buf[MAX_PATH];
  ply_hdl randlist;

  if((ptr=(char*)calloc(playlist.num_rows,playlist.row_size))==NULL)
  { Stop_Server(); exit(1); }
  randlist.alloc_ptr=ptr;
  randlist.num_rows=playlist.num_rows;
  randlist.row_size=playlist.row_size;
  srandom(rand);

  ptr=playlist.alloc_ptr;
  ptr2=randlist.alloc_ptr;
  if(strncmp((char*)ptr,"#",1)==0)
  { sprintf((char*)ptr2,"%s",(char*)ptr);
    sprintf((char*)ptr,"");
    count++;
  }
  for(;count<randlist.num_rows;count++)
  { i=(int) random()/ (RAND_MAX / randlist.num_rows);
    ptr=playlist.alloc_ptr + (playlist.row_size * i);
    ptr2=randlist.alloc_ptr + (randlist.row_size * count);
    while((sprintf((char*)ptr2,"%s",(char*)ptr))==0)
    { i=(int) random()/ (RAND_MAX / randlist.num_rows);
      ptr=playlist.alloc_ptr + (playlist.row_size * i);
    }
    sprintf((char*)ptr,"");
//    printf("%i %i\n",i,count);
  }
  free(playlist.alloc_ptr);
  playlist.alloc_ptr=randlist.alloc_ptr;
  return(playlist);
}

//*******************************************************
int Save_Playlist (playlist,where,randseed)
ply_hdl playlist;
char where[MAX_PATH];
int randseed;
{ long i;
  int fd;
  char buf[MAX_PATH],tbuf[MAX_PATH];
  char *ptr;

  fd=open(where,O_WRONLY|O_TRUNC|O_CREAT,0666);

  for(i=0;i<playlist.num_rows;i++)
  { ptr=playlist.alloc_ptr + (playlist.row_size * i);
//Saving the initial comment containing random seed and playlist origin
    if( (i<1) && (strncmp((char*)ptr,"#",1)==0))
      sprintf(buf,"#%i %s\n",randseed,PLAYLIST);
    else if(i<1) 
    { sprintf(tbuf,"#%i %s\n",randseed,PLAYLIST);
      write(fd,tbuf,strlen(tbuf));
      sprintf(buf,"%s\n",(char*)ptr);
    }
    else sprintf(buf,"%s\n",(char*)ptr);
    write(fd,buf,strlen(buf));
  }
  close(fd);
  return(0);
}

//*******************************************************
void Read_Dir ( name , fd )
char name[MAX_PATH];
int fd;
{ char t_name[MAX_PATH],lt_name[MAX_PATH];
  struct dirent *dp;
  DIR *dfd;
  long len;

  if((dfd=opendir(name))==NULL)
  { printf("Error: Cannot open %s\n",name);
    return; 
  }
  while((dp=readdir(dfd))!=NULL)
  { sprintf(t_name,"%s",dp->d_name);
    if(strcmp(dp->d_name,".")==0 || strcmp(dp->d_name,"..")==0 ||
      (strcmp(strtolower(t_name,lt_name),"no_autoplay"))==0 )
      continue;
    else
    { sprintf(t_name,"%s/%s",name,dp->d_name);
      if(Is_File(t_name)==0)
      { if( (strstr(strtolower(t_name,lt_name),".mp3")) != NULL ) 
        { len=sprintf(t_name,"%s\n",t_name);
          if(write(fd,t_name,strlen(t_name))!=len) 
            printf("Error: Writing to %s\n",t_name);
        }
      }
      else if(Is_Dir(t_name)==0) Read_Dir(t_name,fd); 
    }
  }
}

//*******************************************************
ply_hdl Load_Playlist (name,rand)
char name[MAX_PATH];
int rand;
{ int max=0,l=0,tpfd[2];
  long count=0,i;
  int pl_fd,tfd;
  char *ptr,*ptr2;
  char buf[MAX_PATH],tmps[MAX_PATH];
  char *sp;
  ply_hdl playlist;
  long plt,nft;
  int tmprand;

//generating a unique token for the playlist
//so that a permanent savelist can be made
  path2tok(name,hist_name);
  sprintf(tmps,"%s/%s.msl",TMP_DIR,hist_name);
//if it is an existant file use it for loading 
//and get offset from same file with extension trk
  if(Is_File(tmps)==0)
  { strcpy(name,tmps);
    strcpy(tmps,"\0");
//    rand=0;
    sprintf(buf,"%s/%s.trk",TMP_DIR,hist_name);
    if((tfd=open(buf,O_RDONLY))!=-1) 
    { read(tfd,&START,sizeof(START));
      if(VERBOSE>0) printf("Playlist last track was : %i\n",START);
      if(SFRAME>0) 
      { if((read(tfd,&SFRAME,sizeof(SFRAME)))<=0) SFRAME=0;
        printf("historic: Will try starting at frame no. %i\n",SFRAME);
      }
    }
  }

//if loading a file must first figure out if it is a playlist
//or a single mp3 file
  if(Is_File(name)==0) 
  { if((sp=strrchr(name,'.'))==NULL) exit(1); 
    else 
    { strtolower((char*)++sp,tmps);
      if(strcmp(tmps,"mp3")==0) 
      { if((pl_fd=open(TPL,O_WRONLY|O_CREAT|O_TRUNC,0666))<0) exit(1);
        write(pl_fd,name,strlen(name));
        close(pl_fd);
        pl_fd=open(TPL,O_RDONLY);
      }
//if it is a list must look at the first reserved comment
//to see it list is uptodate and to get the random seed
      else if((strcmp(tmps,"msl")==0)) 
      { pl_fd=open(name,O_RDONLY);
        if(Read_Ln(pl_fd,buf)>0)
        { if(strncmp(buf,"#",1)==0)
          { sprintf(tmps,"%s",&buf[3]);
//if there are any newer files in the tree must regenerate the list
            if(get_newest_mtime(name) < get_newest_mtime(tmps))
            { close(pl_fd);
              unlink(name);
              strcpy(name,tmps);
              strcpy(tmps,strtok(buf," "));
              tmprand=atoi(&tmps[1]);
              if((strcmp(name,SAVELIST)!=0) && (tmprand!=0))
              { RANDOM=tmprand;
                rand=RANDOM;
              }
              else if((strcmp(name,SAVELIST)==0) && (RANDOM==0))
              {  RANDOM=tmprand;
                rand=RANDOM;
              }
            }
            else
            { strcpy(tmps,strtok(buf," "));
              tmprand=atoi(&tmps[1]);
              if(VERBOSE>0)
                printf("Random seed is %i on playlist %s\n",tmprand,name);
              if(tmprand!=0) { rand=0; RANDOM=tmprand; }
            }
//incrementing count coz I already read a line from the file (the comment)
            count++;
          }
//if there was no comment should seek the begining of the file again
          else lseek(pl_fd,0L,0);
        }
      }
      else exit(1);
    }
  }

  if(Is_Dir(name)==0) 
  { if(VERBOSE>0) printf("Loading directory: %s into tempfile %s\n",name,TPL);
    if((pl_fd=open(TPL,O_WRONLY|O_CREAT|O_TRUNC,0666))<0) exit(1);
    
    Read_Dir(name,pl_fd);
    close(pl_fd);
    pl_fd=open(TPL,O_RDONLY);
  }
  
  while((l=Read_Ln(pl_fd,buf))>1) 
  { if((count==0) && (strncmp(buf,"#",1)!=0)) count++;
    count++;
    if(l>max) max=l;
  }
  max++;
  
  playlist.num_rows=count;
  playlist.row_size=max;
  if(VERBOSE>0) printf("%i elements, %i max element length\n",count,max);
  if((ptr=(char*)calloc(count,max))==NULL) 
  { printf("Error: Fatal, Could not allocate memory %i\n",count*max);
    Stop_Server(); 
    exit(1); 
  } 
  playlist.alloc_ptr=ptr;
  count=0;
  lseek(pl_fd,0L,0);
  while((l=Read_Ln(pl_fd,buf))>1)
  { if(VERBOSE>1) printf("%i, %i, %s\n",count,max,buf);
    if((count==0) && (strncmp(buf,"#",1)!=0))
    { ptr2=ptr + (max * count);
      sprintf((char*)ptr2,"#%i %s\n",0,name);
      ++count;
    } 
    ptr2=ptr + (max * count);
    sprintf((char*)ptr2,"%s",buf);
    count++;
  }
  fsync(fileno(stdout));
  close(pl_fd);
//now keeping a history of all played playlists withe their last
//played tracks
  sprintf(tmps,"%s/%s.msl",TMP_DIR,hist_name);
  if((strcmp(name,SAVELIST)!=0) && (Is_File(tmps)!=0))
    Save_Playlist(playlist,tmps,0);
  if(VERBOSE>2) 
  { printf("Saved historic stuff\n");
    fsync(fileno(stdout));
  }
//randomizing playlist
  if(rand>0) playlist=Randomize_Playlist(playlist,rand);
//this saves playlist in last played playlist
//olny if the saved list is not the playlist just loaded
  if(strcmp(name,SAVELIST)!=0) 
    Save_Playlist(playlist,SAVELIST,RANDOM);
  strcpy(tmps,"\0"); 
  if(VERBOSE>0) printf("Loaded playlist %s correctly!\n",name);
  return(playlist);
}

//*******************************************************
int Start_Server ()
{ char b[MAX_PATH],tmpd[MAX_PATH],tbuf[5];
  char mpg[255],snmpg[255],mpgo[10],mpgfp[30]="-q";
  extern int errno;

//Creating directory for temporary files, command pipe, status pipe and 
//current playlist and playfile
  if (mkdir((char*)TMP_DIR,0777) != 0)
  { a1=errno;
    sprintf(b,"mkdir %s",TMP_DIR,0777); 
    if(Handle_Error(a1,(char*)b) != 0) exit(1);
  } 
  
  if (mkfifo(CMD_PIPE,0666) != 0)
  { a1=errno;
    sprintf(b,"mkfifo %s",CMD_PIPE);
    if(Handle_Error(a1,(char*)b) != 0) exit(1);
  }
  
  if (mkfifo(STAT_PIPE,0666) != 0)
  { a1=errno;
    sprintf(b,"mkfifo %s",STAT_PIPE);
    if(Handle_Error(a1,(char*)b) != 0) exit(1);
  }

#ifdef MPG
    strcpy(mpg,MPG);
#else
    strcpy(mpg,"/usr/bin/mpg123");
#endif
    strcpy(snmpg,(char*)basename(mpg));
    if(strcmp(snmpg,"mpg321")==0)
    { strcpy(mpgo,"oss");
#ifdef SKIP321
      sprintf(mpgfp,"--skip-printing-frames=%i",SKIP321);
#else
      strcpy(mpgfp,"--skip-printing-frames=15");
#endif
    }
    else
    { strcpy(mpgo,"l");
    }

//Checking if mpg software is present
    int mtfd;
    if((mtfd=open(mpg,O_RDONLY)) < 0)
    { printf("%s was not found ... aborting\n", mpg);
      exit(1);
    }
    close(mtfd);
//Starting mpg123 in remote control mode
  if ((pid=fork()) == 0)
  { if (VERBOSE>0) printf("Child is opening pipes ... ");
//mpg123 should only read from cmd_fd[0] 
//and write any output to stat_fd[1]
//now opening named pipes in normal mode
//this will sleep until father opens the other side of the pipes
    cmd_fd[0]=open(CMD_PIPE,O_RDONLY);
    stat_fd[1]=open(STAT_PIPE,O_WRONLY);
    if (VERBOSE>0) printf("Done.\n");
//now wrapping stdin,out,err on pipes
    dup2(cmd_fd[0],fileno(stdin));
    dup2(stat_fd[1],fileno(stdout));
    dup2(stat_fd[1],fileno(stderr));
//Some audio chipsets fave design flaws
//it may be necessary to reopen audio stream every time a new file is loaded
//this may produce a click each time but it may solve some othe problems
//add --reopen amongst options
    if(BUFFER>0)
    { sprintf(tbuf,"%i",BUFFER);
      execl(mpg,snmpg,"-o",mpgo,"-a","/dev/dsp",mpgfp,
        "-b",tbuf,"-q","-y","-R","a",0);
    }
    else
      execl(mpg,snmpg,"-o",mpgo,"-a","/dev/dsp",mpgfp,
        "-q","-y","-R","a",0);
    fsync(stat_fd[1]);
    sync();
    sleep(4);
    close(cmd_fd[0]);
    close(stat_fd[1]);
    Stop_Server();
  } 
  else if (pid < 0)
  { printf("Fork failed !\n");
    exit(1);
  }
  else
  { if(VERBOSE>0) 
    { printf("Forked mp3 player, pid is: %i \n",pid);
      printf("Manager is opening pipes ... ");
    }
//Opening the other side of the pipes that child is trying to open
//this will wake up the child's sleeping opens
//order is important as child is sleeping on it's firts open
    cmd_fd[1]=open(CMD_PIPE,O_WRONLY);
    stat_fd[0]=open(STAT_PIPE,O_RDONLY); 
    stat_fd[1]=open(STAT_PIPE,O_WRONLY);
    if (VERBOSE>0) printf("Done.\n");   
  }
  return 0;
}

//*******************************************************
int Stop_Server ()
{ save_info();
  if (VERBOSE>0) printf("Terminating mp3 player ... ");
//terminating mpg123
  write(cmd_fd[1],"QUIT\n",5); 
//making sure an eventual Manage_file gets quit request
  write(stat_fd[1],"#E QUIT\n",8);
//waiting for mpg123 to quit and close it's side of pipes
  while( pid1 = waitpid(pid, &status, 0) > 0 ) { }
  if (VERBOSE>0) printf("Done!\n");
//Closing side of pipes opened by this process 
#ifdef SPC
  kill(spc_pid,9);
#endif
  kill(ec_pid,9);
  close(cmd_fd[1]);
  close(stat_fd[0]);
  close(stat_fd[1]);
  return 0;
}

//*******************************************************
int Load_File (name)
char name[MAX_PATH];
{ int count;
  char new[MAX_PATH + 5];

  if(Is_File(name)!=0) return(-1);
  
  count=sprintf(new,"LOAD %s\n",name);
  if(write(cmd_fd[1],new,count)==count)
  { if(SFRAME>0)
    { sleep64(1);
      count=sprintf(new,"JUMP %i\n",SFRAME);
      SFRAME=(long)0;
      if(write(cmd_fd[1],new,count)==count) return(0);
    }
    return(0);
  }
  else
    return(-1);
}

//*******************************************************
//This function returnes:
//x; x<=-10 or x>=10 skip x tracks "#E SKIP X" 
//-1 it unrecoverable error or quit request arrived "#E QUIT"
//0 track finished correctly or force finish track request arrived "@P 0"
//1 if restart from begining of playlist request arrived "#E BEGIN"
//2 if restart playlist of current file "#E RESTART"
//3 if reload previous track request arrived "#E PREV"
//4 if randomize playlist "#E RAND"
//5 if repeat playback "#E REPEAT" will try to do w/o actually returning
//6 if load new playlist arrived "#L playlist"
//7 if pause playback arrived "#E PAUSE" will try to do w/o actually returning
//8 if reverse playback arrived "#E BACKPPLAY" `will try to do w/o returning
//9 #E SHOWID3TAG will do w/o returning
int Manage_File (offset,playlist,trackno)
long offset;
ply_hdl playlist;
long trackno;
{ char buf[MAX_PATH],buf2[MAX_PATH];
  char id3tag[MAX_PATH],ttag[MAX_PATH];
  char title[30],artist[30],album[30],year[4],comment[30];
  char *spt,*spt2;
  int ret;
  

//  while(readmsg(PROJ_NAME,buf2)!=-1) { sleep(1); }
  while(readmsg(PROJ_NAME,buf2)!=-1) { }
  sprintf(buf2,"%i: %s",trackno,(char*)offset);
  sendmsg(PROJ_NAME,buf2);
  if (VERBOSE>0) printf("Loading: %s @ frame no %i\n",(char*)offset,SFRAME);
//Pausing 1/4 second before loading next song
//Introduced this dew to a dma problem on my audio card
  sleep64(16);
  if(Load_File((char*)offset)!= 0)
  { if (VERBOSE>0) printf("Problem loading %s\n",(char*)offset);
    sendmsg(PROJ_NAME,"Could not load: skipping to next track!");
    return(0);
  }
  while(1)
  { Read_Ln(stat_fd[0],buf);
//Player release information at startup
    if( strncmp(buf,"@R ",3)==0 ) { if(VERBOSE>1) printf("%s\n",buf); }
//ID3 tag of loaded file or just filename if no ID3 tags present
    else if( strncmp(buf,"@I ",3)==0 ) 
    { if(VERBOSE>1) printf("%s\n",buf); 
      sprintf(ttag,"%s",&buf[3]);
      cut_space(id3tag,ttag);
    }
//Status message after loading a song (stream info)
    else if( strncmp(buf,"@S ",3)==0 ) { if(VERBOSE>1) printf("%s\n",buf); }
//playback stream info 
    else if( strncmp(buf,"@F ",3)==0 ) 
    { if(VERBOSE>3) printf("%s\n",buf);
      spt=strchr(buf,' ');
      spt2=strtok(&spt[1]," ");
      frame=atol(spt2); 
      cur_track=trackno;
    }    
//Playing status
    else if( strncmp(buf,"@P 0",4)==0 )
    { if(VERBOSE>1) printf("finished playing file\n");
      break;
    }
    else if( strncmp(buf,"@P ",3)==0 ) { if(VERBOSE>1) printf("%s\n",buf); }
//An error occured
    else if(strncmp(buf,"@E "       ,3 )==0) printf("Error $=%s\n",buf); 
//cm3proto commands
    else if(strncmp(buf,"#E SKIP"   ,7 )==0)
    { ret=atoi(&buf[7]);
      if (VERBOSE>1) printf("Skipping %i tracks!",ret);
      if((ret <= -10) || (ret >= 10))           return(ret); 
      else continue;
    }
    else if(strncmp(buf,"#E QUIT"   ,7 )==0) 
    { printf("Received QUIT request! \n"); return(-1); }
    else if(strncmp(buf,"#E BEGIN"  ,8 )==0)    return(1);
    else if(strncmp(buf,"#E RESTART",10)==0)    return(2);
    else if(strncmp(buf,"#E PREV"   ,7 )==0)    return(3); 
    else if(strncmp(buf,"#E RAND"   ,7 )==0)    return(4);
    else if(strncmp(buf,"#E REPEAT" ,9 )==0)
    { if (REPEAT==0) 
      { REPEAT=1; sendmsg(PROJ_NAME,"Repeat ON"); } 
      else { REPEAT=0; sendmsg(PROJ_NAME,"Repeat OFF"); } 
    }
    else if(strncmp(buf,"#L "       ,3 )==0)
    { sprintf(PLAYLIST,"%s",&buf[3]);      return(6); } 
    else if(strncmp(buf,"#E PAUSE"  ,8 )==0)   
      write(cmd_fd[1],"PAUSE\n",6);
    else if(strncmp(buf,"#E BACKPLAY",11)==0)
    { if(BACKPLAY==0) { BACKPLAY=1; sendmsg(PROJ_NAME,"Reverse ON"); }
      else { BACKPLAY=0; sendmsg(PROJ_NAME,"Reverse OFF"); }
    }
    else if(strncmp(buf,"#E SHOWID3TAG",13)==0) 
    { Read_Ln(stat_fd[0],buf);
      spt2=strrchr(buf,' ');
      spt2[0]=10; 
      ++spt2;
      spt=strrchr(buf,' ');
      spt[0]=10; 
      ++spt;
      sprintf(ttag,"%s\nPlayed %i of %i sec total playtime.",
        id3tag,atoi(spt),atoi(spt2)+atoi(spt));
      sendmsg(PROJ_NAME,ttag);
    }
//Null OP parsed ... this was gust to make sure a new line is in pipe
//w/o having an empty line
    else if(strncmp(buf,"NOP",3)==0) { if(VERBOSE>1) printf("%s\n",buf); }
//this in not anything ither in cm3proto or player info
//it is probably something written to player stderr
    else { printf("Unparsed: %s\n",buf); fsync(fileno(stdout)); }
  }
  return(0);
}

//*******************************************************
int Manager ( playlist, start_track )
ply_hdl playlist;
long start_track;
{ long i,wcmd=0;
  int tfd;
  int stat,tr;
  char buf[MAX_PATH],tmps[MAX_PATH];
  ply_hdl *tply;

//This prevents from starting out of range
//It may happen if you delete some entries w/o updating 
//lat played track 
  while(start_track<0) { start_track=start_track + playlist.num_rows; }
  while(start_track>=playlist.num_rows) 
    { start_track=start_track - playlist.num_rows; }
  if((start_track<0) || (start_track>=playlist.num_rows)) start_track=0;

  i=start_track;
  while((i<playlist.num_rows) && (i>=0))
  { if(VERBOSE>1) printf("Trach no. %i\n",i); 
//Leaving out any # comments
    if(strncmp((char*)(playlist.alloc_ptr+(playlist.row_size*i)),"#",1)==0) 
    { if(BACKPLAY==0) ++i; 
      else --i;
      if(REPEAT != 0) 
      { while(i<0) { i=i + playlist.num_rows; }
        while(i>=playlist.num_rows) { i=i - playlist.num_rows; }
      }
      continue; 
    }

//saving track and frame information
    cur_track=i;
    save_info();
    stat=Manage_File((long)(playlist.alloc_ptr+(playlist.row_size*i)),
      playlist,i);
    if(BACKPLAY==0) ++i; 
    else --i;
    if (stat >= 10) i = i + stat -1;
    if (stat <= -10) i = i + stat -1;
    else if ( stat == -1 ) break;
    else if ( stat == 1 ) i=0;
    else if ( stat == 2 ) --i;
    else if ( stat == 3 ) i=i - 2;
    else if ( stat == 4 ) 
    { if(RANDOM==0) 
      { RANDOM=1;
        if(VERBOSE>0) printf("Randomizing playlist %s\n",PLAYLIST);
        sendmsg(PROJ_NAME,"Random ON");
      }
      else 
      { RANDOM=0;
        if(VERBOSE>0) printf("Unrandomizing playlist %s\n",PLAYLIST);
        sendmsg(PROJ_NAME,"Random OFF");
      }
      tr=RANDOM; 
      strcpy(tmps,PLAYLIST);
      playlist=Load_Playlist(tmps,tr); 
    }
    else if ( stat == 5 ) 
    { if (REPEAT==0) REPEAT=1; 
      else REPEAT=0;  
    }
    else if ( stat == 6 ) 
    { 
      if ((Is_File(PLAYLIST)==0)|| (Is_Dir(PLAYLIST)==0))
      { START=0;
        if (VERBOSE>0) printf("Loading playlist: %s\n",PLAYLIST);
        strcpy(tbuf,PLAYLIST);
        tr=RANDOM;
        free(playlist.alloc_ptr);
        playlist.num_rows=0;
        playlist.row_size=0;
        playlist=Load_Playlist(tbuf,tr);
        i=START;
      }
    }
    else if ( stat == 7 ) write(cmd_fd[1],"PAUSE\n",6);
    wcmd=0;
    if(REPEAT != 0) 
    { while(i<0) { i=i + playlist.num_rows; }
      while(i>=playlist.num_rows) { i=i - playlist.num_rows; }
    }
//If you get to this point then you finished playing a non repeat list
//will stop here and wait for a command
    if((i>=playlist.num_rows) || (i<0))
    { write(cmd_fd[1],"S\n",2);
      while(Read_Ln(stat_fd[0],tmps)==0) 
      { if(strncmp(tmps,"@P 0",4)==0) break; }
      sendmsg(PROJ_NAME,"Waiting .... Tell me what to play next!!!!!");
      printf("Waiting .... Tell me what to play next!!!!!\n");
      while(1)
      { Read_Ln(stat_fd[0],tmps);
        if(strncmp(tmps,"#E QUIT"   , 7)==0) break;
        else if(strncmp(tmps,"#E BEGIN"  , 8)==0) { i=0; break; }
        else if(strncmp(tmps,"#E RESTART",10)==0) break; 
        else if(strncmp(tmps,"#E PREV"   , 7)==0) { i--; break; }
        else if(strncmp(tmps,"#E REPEAT" , 9)==0) 
        { REPEAT=1; i=0; sendmsg(PROJ_NAME,"Repeat ON"); break; }
        else if(strncmp(tmps,"#L "       , 3)==0) 
        { sprintf(PLAYLIST,"%s",&tmps[3]);
          if ((Is_File(PLAYLIST)==0)|| (Is_Dir(PLAYLIST)==0))
          { START=0;
            printf("Loading playlist: %s\n",PLAYLIST);
            strcpy(tbuf,PLAYLIST);
            tr=RANDOM;
            free(playlist.alloc_ptr);
            playlist.num_rows=0;
            playlist.row_size=0;
            playlist=Load_Playlist(tbuf,tr);
            i= START;
            break;
          }
        }
        else if(strncmp(tmps,"NOP",3)!=0)
        sendmsg(PROJ_NAME,
          "It might be a good idea to load a new playlist!!!");  
        printf("It might be a good idea to load a new playlist!!!\n");
      }
      if(VERBOSE>1) printf("Received a command\n");
//      sleep(2);
    }
  }
//  fsync(fileno(stdout));
  return(0);
}

//*******************************************************
int main (int argc, char **argv)
{ long offset=-1,tlong;
//I need to keep track of the actual requested random capability
//and the need to suppress random playback when loading last saved playlist
  int random;
  int tfd,tfd2;
  char buf[MAX_PATH],buf2[MAX_PATH];
  char *ptr;
  int ti;
  char ini_file[MAX_PATH];
  char **mixer_str[4],mstr[80],pstr[80];
  
  signal(15, sig_quit_handler);
  signal(2, sig_quit_handler);
  umask(0000);

//Looking for ini file first in ~/.cm3s.ini then in /etc/cm3s.ini
  sprintf(ini_file,"%s/.%s",getenv("HOME"),INI_FILE);
  if(Is_File(ini_file)!=0) sprintf(ini_file,"/etc/%s",INI_FILE);
  if(Is_File(ini_file)==0) 
  { printf("Reading options from: %s\n",ini_file);
    get_ini_options(ini_file);
  } 
  else printf("Could not find any configuration files\n");

  if(argc>1) get_options(argc,argv);

//if there is mixer support will set the volumes
#ifdef MIXER
  mixer_str[0]=(char**)"vol";
  mixer_str[1]=(char**)&MASTER_VOL[0];
  mixer_str[2]=(char**)"pcm";
  mixer_str[3]=(char**)&PCM_VOL[0];
  cmd_mixer(4,mixer_str);
#endif

  if(NEXT==1) SFRAME=(long)0;
  random=RANDOM;
  sprintf(buf,"%s/%s",TMP_DIR,STAT_PIPE);
  strcpy(STAT_PIPE,buf);
  sprintf(buf,"%s/%s",TMP_DIR,CMD_PIPE);
  strcpy(CMD_PIPE,buf);
  sprintf(buf,"%s/%s",TMP_DIR,SETTINGS);
  strcpy(SETTINGS,buf);
  if(strcmp(SAVELIST,PLAYLIST)==0)
  { sprintf(buf,"%s/%s",TMP_DIR,PLAYLIST);
    strcpy(PLAYLIST,buf);
  }
  sprintf(buf,"%s/%s",TMP_DIR,SAVELIST);
  strcpy(SAVELIST,buf);
  sprintf(buf,"%s/%s",TMP_DIR,TPL);
  strcpy(TPL,buf);
  sprintf(buf,"%s/%s",TMP_DIR,OFFSET);
  strcpy(OFFSET,buf);
  if(strcmp(LOG,"log")==0)
  { sprintf(buf,"%s/%s",TMP_DIR,LOG);
    strcpy(LOG,buf);
  }

#ifdef FPFK
  if(get_ug(FPFK,ugids)!=0) 
    printf("Could not get uid and gid of %s\n",FPFK);
#else
  if(get_ug("/usr/local/bin/cm3s",queue)!=0)
    printf("Could not get uid and gid of /usr/local/bin/cm3s");
#endif

  strcpy((char*)progname,(char*)basename(argv[0]));
//checking any previous instances of cm3s (and children)
//along with anything that uses /dev/dsp
  check();

//forking the spc stuff here if configure found a compatible system
#ifdef SPC
if((spc_pid = fork()) == 0)
{
//setting ioperms on the parallel port device (need to do this before dropping
//privileges
  set_ioperms(LPT_IO);
//dropping privileges (the spc will only work if you start up from root)
  set_rugid(ugids);
//checking if the spc is connected
  data_out(255,LPT_IO);
  sleep(1);
  if(check_hw(LPT_IO) != 0) 
  { if(VERBOSE>0) printf("No SPc hardware was found!!!\n");
    exit(0);
  }
  data_out(1,LPT_IO);
  spc_manager(LPT_IO);
  data_out(0,LPT_IO);
  exit(0);
}
#endif

//If running cm3s_boot will remain root as a bug workaround
  if(strcmp(progname,"cm3s_boot")==0) 
    printf("Reamining as root for autostart\n");
  else
  { if((ti=set_rugid(ugids))!=0) 
      printf("Warning: could not change user correctly! %i\n",ti);
  }

  if(VERBOSE>2) printf("changed user\n");  

  if(BEDAEMON==1)
  { if((ti=be_daemon())==-1) printf("Warning %i: could not go in background!\n",ti);
    else
    { if((tfd=open(LOG,O_CREAT|O_WRONLY|O_TRUNC,0666))==-1) 
      { tfd=open("/dev/null",O_WRONLY);
        printf("Warning: could not open log file: %s\n",LOG);
      }
      tfd2=open("/dev/null",O_RDONLY);
      if((ti=realloc_stdioe(tfd2,tfd,tfd))!=0) 
        if(VERBOSE>1)
          printf("Warning: error %i reassigning standart in/out/err\n",ti); 
    }
  }
 
//uncomment this only if you get a coutinuous low freq buz
//  if((ec_pid=fork())==0) eat_cpu();
 
  msg_del_queue(PROJ_NAME);

  Start_Server();
  if (pid != 0 ) 
  { if(VERBOSE>2) printf("started server\n");
//handling playlist importance
//PLAYLIST mus be ither a file or a directory
    if(Is_File(PLAYLIST)!=0 && Is_Dir(PLAYLIST)!=0)
//if it is not I will examine the last savelist which must be a file
    { if(Is_File(SAVELIST)!=0) 
//it it is not then I'll examine the default mp3 directory and
//try to use that as a playlist
      { if(Is_Dir(MP3_DIR)==0) strcpy(PLAYLIST,MP3_DIR);
        else
//if that fails too I'll give up ... cannot figure out what you want to play
        { printf("Could not load any playlist ... quitting.\n");
          Stop_Server();
          exit(1);
        }
      }
//at this point you are loading from a saved list
      else random=0;
    }
//at this point you are loading last savelist 
//so you should not randomize it again
    else if(strcmp(PLAYLIST,SAVELIST)==0) random=0;
 
//Handling START point importance
//Will only examine the content of OFFSET if
//1) it is a file
//2) The playlist that is being loaded is the last saved plalist
    if(Is_File(OFFSET)==0 && strcmp(PLAYLIST,SAVELIST)==0)
    { tfd=open(OFFSET,O_RDONLY);
      read(tfd,&offset,sizeof(offset));
      if(SFRAME>0) 
      { if((read(tfd,&SFRAME,sizeof(SFRAME)))<=0) SFRAME=0;
        printf("last: Will try starting from frame no. %i\n",SFRAME);
      }
    }
//So now offset contains the eventual value got from last playlist or -1
//while START has the value that getopt passed or -1
//playlist loadin will overwrite START with any historic stuff
//so I'll save a copy in tlong
    tlong=START;
//now loading the playlist into memory structure 
    strcpy(tbuf,PLAYLIST);
    playlist=Load_Playlist(tbuf,random);
    if(VERBOSE>2) printf("correclty loaded and saved histoic list\n");
//now I'm recalling the saved startpoint before playlist loading
//toffset is from command line, START is from the histotic, offset is from last
//the leas important is the one saved il last
//then the one from historic
//then command line
    if(START!=-1) offset=START;
    if(tlong!=-1) offset=tlong;
    if(offset!=-1) START=offset; 
    else START=0;

//Starting from start + next if next is > 0
    if(NEXT>0) START=START + NEXT;

    if(VERBOSE>0) printf("Staring at track %i\n",START);
    printf("You should now start client software to manipulate cm3s!!!!!");
//now starting to play playlist
    Manager(playlist,START);
    if(VERBOSE>2) printf("started playback ... all done!!!\n");
//if code gets to this point ither something had a fatal error
//or got a quit request
    fsync(fileno(stdout));
    Stop_Server();
    free(playlist.alloc_ptr);
    msg_del_queue(PROJ_NAME);
  }
//waiting for any childs to exit
  while( pid1 = waitpid(pid, &status, 0) > 0 ) { }
}

