/*
MP3 STATION

A plain text Linux/Unix toolset for managing mp3 playback with as little
user interaction as possible.  It was built to handle mp3 playback in a car.
Copyright (C) 2003 David Rao <louigi600@yahoo.it>

This file is part of MP3 STATION.

MP3 STATION is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MP3 STATION is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MP3 STATION; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*
This is an ini like configuration file reader and manager.
ini file comments begin with #
Blanks and tabs will be ignored.
Blank lines should be ignored.

A parameter line looks like this:
parameter_name = parameter_value

Parameters and names are stored in a name ordered tree structure
see ini_ent structure definition

iealloc allocates memory for a new mapameter node and returnes pointer 
to allocated memory (a bit like malloc)

ins_ent inserts a new element in the tree

read_ini_file loads parameters from the given file

get_ent_val returnes value on parameter whose name is given
in arguments (if not present it returnes a null string) 

Look at the commented main if you need help on how to use this
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


//Typedefs
//***************************************************************************
//pointer to elements in ini parameter tree
typedef struct ini_ent_node *ini_ent_ptr;

//ini parameter entry tree node structure 
//with double pointers for up/down navigation if necessary
//but unused for the moment
typedef struct ini_ent_node
{ char ent_name[80];
  char ent_val[176];
  ini_ent_ptr up;
  ini_ent_ptr left;
  ini_ent_ptr right;
} ini_ent; 

//External variables
//***************************************************************************
extern char INI_FILE[];
extern char MP3_DIR[];
extern char TMP_DIR[];
extern char STAT_PIPE[];
extern char CMD_PIPE[];
extern char SETTINGS[];
extern char PLAYLIST[];
extern char LOG[];
extern char TPL[];
extern char OFFSET[];
extern char VERSION[];
extern int BEDAEMON;
extern int RANDOM;
extern int BACKPLAY;
extern int REPEAT;
extern long START;
extern long SFRAME;
extern int NEXT;
extern int BUFFER;
extern int VERBOSE;
extern char MASTER_VOL[];
extern char PCM_VOL[];
extern int LPT_IO;

//Global variables
//***************************************************************************
ini_ent_ptr tree_root;

//Function declarations
//***************************************************************************
ini_ent_ptr iealloc();
int ins_ent( char name[80], char val[176], ini_ent_ptr up_ent);
char *get_ent_val(char name[80], char val[176]);
int read_ini_file(char name[1024]);
void show_ent(ini_ent_ptr ent);
void free_ent(ini_ent_ptr ent);
void ptvisit(ini_ent_ptr up_ent, void (*func)(ini_ent_ptr));
void get_ini_options(char name[1024]);

//Function definitions
//***************************************************************************
//allocate memory for ini entry node
ini_ent_ptr iealloc()
{ return((ini_ent_ptr) malloc(sizeof(ini_ent)));
}

//Recursive node insert into ini entry tree
ins_ent(name,val,ent)
char name[80];
char val[176];
ini_ent_ptr ent;
{ if(ent->up == NULL)
//Here we are inserting into tree root, otherwise tree_root.up should
//point to itself
  { strcpy(ent->ent_name,name);
    strcpy(ent->ent_val,val);
    ent->up=ent;
  }
  if(strcmp(ent->ent_name,name)==0) strcpy(ent->ent_val,val);
  else if(strcmp(ent->ent_name,name)<0)
  { if(ent->right != NULL) ins_ent(name,val,ent->right);
    else
    { ini_ent_ptr new_ent=iealloc();
      ent->right=new_ent;
      new_ent->right=NULL;
      new_ent->left=NULL;
      new_ent->up=ent;
      strcpy(new_ent->ent_name,name);
      strcpy(new_ent->ent_val,val);
    } 
  }
  else
  { if(ent->left != NULL) ins_ent(name,val,ent->left);
    else
    { ini_ent_ptr new_ent=iealloc();
      ent->left=new_ent;
      new_ent->right=NULL;
      new_ent->left=NULL;
      new_ent->up=ent;
      strcpy(new_ent->ent_name,name);
      strcpy(new_ent->ent_val,val);
    }
  }
}

//get a value from ini entry tree with a specific entry name 
char *get_ent_val(name,val)
char name[80];
char val[176];
{ ini_ent_ptr ent=tree_root;
  int cmp=0;

  while( (cmp=strcmp(ent->ent_name,name))!=0 )
  { if((cmp<0) && (ent->right!=NULL)) ent=ent->right;
    else if((cmp<0) && (ent->right==NULL)) break;
    else if((cmp>0) && (ent->left!=NULL)) ent=ent->left;
    else  break;
  }
  if(strcmp(ent->ent_name,name)==0) 
  { strcpy(val,ent->ent_val); 
    return(val);
  }
  else return("");
}

//read ini file and load values into ini entry tree
int read_ini_file(name)
char name[1024];
{ int fd;
  char buf[256],buf2[256],tname[80],tval[176];
  int i,j;

  tree_root=iealloc();
  strcpy(tree_root->ent_name,"");
  strcpy(tree_root->ent_val,"");
  tree_root->up=NULL;
  tree_root->left=NULL;
  tree_root->right=NULL;
  fd=open(name,O_RDONLY);
  while((i=read_ln(fd,buf))!=0)
  { rm_space(buf2,buf);
    if((strncmp("#",buf2,1)!=0) && (strlen(buf2)>1)) 
    { //printf("%s\n",buf2);
      strcpy(tval,(char*)strchr(buf2,'='));
      strcpy(tname,buf2);
      strtok(tname,"=");
      ins_ent(tname,&tval[1],tree_root); 
    }
  }
  
  close(fd); 
}

void free_ent(ent)
ini_ent_ptr ent;
{ free(ent);
}

void show_ent(ent)
ini_ent_ptr ent;
{ printf("%s, %s\n",ent->ent_name, ent->ent_val);
}

//Post order tree visit and execute func at each node/leaf
void ptvisit(ent, func)
ini_ent_ptr ent;
void (*func)(ini_ent_ptr);
{ if(ent->up==NULL) return; //This avoids visiting an empty tree root
  if(ent->left!=NULL) ptvisit(ent->left,func);
  if(ent->right!=NULL) ptvisit(ent->right,func);
  (*func)(ent);
}

void get_ini_options(name)
char name[1024];
{ char e1[176],e2[176];

  read_ini_file(name);
//  ptvisit(tree_root,show_ent);


  get_ent_val("mp3_base_directory",e1);
  if(strlen(e1)>1) strcpy(MP3_DIR,e1); 
  strcpy(e1,"");

  get_ent_val("temporary_directory",e1);
  if(strlen(e1)>1) strcpy(TMP_DIR,e1);
  strcpy(e1,"");
//STAT_PIPE
//CMD_PIPE
//SETTINGS

  get_ent_val("default_playlist",e1);
  if(strlen(e1)>1) strcpy(PLAYLIST,e1);
  strcpy(e1,"");

  get_ent_val("log_file",e1);
  if(strlen(e1)>1) strcpy(LOG,e1);
  strcpy(e1,"");
//TPL
//OFFSET

  get_ent_val("run_as_daemon",e1);
  if((strlen(e1)>1) && (strcmp(e1,"false")!=0)) BEDAEMON=1;
  strcpy(e1,"");

  get_ent_val("play_in_random_mode",e1);
  if((strlen(e1)>1) && (strcmp(e1,"false")!=0)) RANDOM=1;
  strcpy(e1,"");

  get_ent_val("reverse_playlist_order",e1);
  if((strlen(e1)>1) && (strcmp(e1,"false")!=0)) BACKPLAY=1;
  strcpy(e1,"");

  get_ent_val("repeat_playlist",e1);
  if((strlen(e1)>1) && (strcmp(e1,"false")!=0)) REPEAT=1;
  strcpy(e1,"");
//START

  get_ent_val("save_frame",e1);
  if((strlen(e1)>1) && (strcmp(e1,"false")!=0)) SFRAME=(long)1;
  strcpy(e1,"");

  get_ent_val("start_next",e1);
  if((strlen(e1)>1) && (strcmp(e1,"false")!=0)) NEXT=1;
  strcpy(e1,"");

  get_ent_val("audio_buffer",e1);
  if(strlen(e1)>1) BUFFER=atoi(e1);
  strcpy(e1,"");
//VERBOSE

  get_ent_val("master_vol",e1);
  if(strlen(e1)>1) strcpy(MASTER_VOL,e1);
  strcpy(e1,"");

  get_ent_val("pcm_vol",e1);
  if(strlen(e1)>1) strcpy(PCM_VOL,e1);
  strcpy(e1,"");

  get_ent_val("spc_port",e1);
  if(strlen(e1)>1) LPT_IO=atox(e1);
  strcpy(e1,"");

  ptvisit(tree_root,free_ent);
}

//main ()
//{ char buf[256];
//
//  read_ini_file("/tmp/test.ini");
//  printf("----------------------\n");
//  printf("%s\n",get_ent_val("aname",buf));
//  printf("----------------------\n");
//  ptvisit(tree_root,show_ent);
//  ptvisit(tree_root,free_ent);
//}
