/*
MP3 STATION

A plain text Linux/Unix toolset for managing mp3 playback with as little
user interaction as possible.  It was built to handle mp3 playback in a car.
Copyright (C) 2003 David Rao <louigi600@yahoo.it>

This file is part of MP3 STATION.
This is a crippled and modified version of rexima to suit the 
client program needs.
  rexima 1.2 - a curses-based (and command-line) mixer for Linux.
  Copyright (C) 1996-2002 Russell Marks.
Many thanks to Russell Marks.

MP3 STATION is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MP3 STATION is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MP3 STATION; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <ncurses.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>


#define DEV_Y_START	1
#define DEV_X_START	0
#define DEV_X_DEVEND	(DEV_X_START+9)
#define DEV_X_BAR	(DEV_X_DEVEND+2)
#define DEV_X_PCNT	(DEV_X_BAR+53)
#define DEV_X_REC	(DEV_X_PCNT+5)

typedef char help[][80];

extern WINDOW *w,*w2,*w3,*w4;

help mixer_help={
  "e,esc    : Exit mixer and return to mani menu",
  "arrorL,- : Turn vol down        , arrowR,+ : Turn vol up",
  "arrowU,k : Change selection up  , arrowD,j : Change selection down",
  ""
};
char dev_labels[SOUND_MIXER_NRDEVICES][80]=SOUND_DEVICE_LABELS;
char dev_names [SOUND_MIXER_NRDEVICES][80]=SOUND_DEVICE_NAMES;
int dev_line[SOUND_MIXER_NRDEVICES];

//***********************************************************
void die(char *str)
{ fprintf(stderr,"mixer: couldn't %s.\n",str);
  exit(1);
}

//***********************************************************
void init(mixfd,existmask,canrecmask,isrecmask,stereomask)
int *mixfd,*existmask,*canrecmask,*isrecmask,*stereomask;
{ if(((*mixfd)=open("/dev/mixer",O_RDWR))<0) die("open /dev/mixer");
  if(ioctl(*mixfd,SOUND_MIXER_READ_DEVMASK,existmask)==-1) die("ioctl");
  if(ioctl(*mixfd,SOUND_MIXER_READ_RECMASK,canrecmask)==-1) die("ioctl");
  if(ioctl(*mixfd,SOUND_MIXER_READ_RECSRC,isrecmask)==-1) die("ioctl");
/* this looks like a `recent' addition [in 1996 maybe ;-)] so be lenient */
  if(ioctl(*mixfd,SOUND_MIXER_READ_STEREODEVS,stereomask)==-1) *stereomask=0;
}

//***********************************************************
void init_term()
{ wclear(w2); wrefresh(w2);
}

//***********************************************************
void uninit(int mixfd)
{
  wclear(w2); wrefresh(w2);
  close(mixfd);
}

//***********************************************************
void drawsel(int new,int old)
{
  if(new!=old)
  { if(old>=0) mvwaddstr(w2,DEV_Y_START+dev_line[old],DEV_X_DEVEND,"  ");
    if(new>=0) mvwaddstr(w2,DEV_Y_START+dev_line[new],DEV_X_DEVEND,"*");
  }
/* this'll be the last thing before a refresh, so... */
  move(DEV_Y_START+dev_line[new],0);
}

//***********************************************************
int mixer_getlevel_stereo(int mixfd,int dev)
{ int level=0;

  ioctl(mixfd,MIXER_READ(dev),&level);
  return(level);
}

//***********************************************************
int mixer_getlevel(int mixfd,int dev)
{ return(mixer_getlevel_stereo(mixfd,dev)&255);
}

//***********************************************************
void mixer_setlevel_stereo(int mixfd,int dev,int left,int right)
{ left+=256*right;
  ioctl(mixfd,MIXER_WRITE(dev),&left);
}

//***********************************************************
void mixer_setlevel(int mixfd,int dev,int level)
{ mixer_setlevel_stereo(mixfd,dev,level,level);
}

//***********************************************************
void mixer_change(int mixfd,int dev,int add)
{ int level=mixer_getlevel(mixfd,dev)+add;

  if(level<0) level=0;
  if(level>100) level=100;
  mixer_setlevel(mixfd,dev,level);
}

//***********************************************************
void mixer_rectoggle(int mixfd,int dev,int *isrecmask)
{ (*isrecmask)^=(1<<dev);
  ioctl(mixfd,SOUND_MIXER_WRITE_RECSRC,isrecmask);
}

//***********************************************************
void mixer_recset(int mixfd,int dev,int *isrecmask,int on)
{ if(on)
  (*isrecmask)|= (1<<dev);
  else
  (*isrecmask)&=~(1<<dev);
  ioctl(mixfd,SOUND_MIXER_WRITE_RECSRC,isrecmask);
}

//***********************************************************
void drawlevel(int dev,int level)
{ char buf[60];
  int f;

/* sanity check */
  if(level<0) level=0;
  if(level>100) level=100;

  memset(buf,'=',51);
  buf[51]=0;
  buf[level/2]='|';
  for(f=level/2+1;f<51;f++)
    buf[f]=' ';

  mvwaddstr(w2,DEV_Y_START+dev_line[dev],DEV_X_BAR+1,buf);

  sprintf(buf,"%3d%%",level);
  mvwaddstr(w2,DEV_Y_START+dev_line[dev],DEV_X_PCNT,buf);
}

//***********************************************************
void setupframe(int existmask,int canrecmask,int *firstdevp,int *lastdevp)
{ int f;
  int offset=0;

  mvwaddstr(w2,DEV_Y_START-1,DEV_X_BAR+1,
    "min   .     .     .      |      .     .     .   max");
/* we know existmask is non-zero by now (see main()) */
  for(f=0;f<SOUND_MIXER_NRDEVICES;f++)
  { if(existmask&(1<<f))
    { if(offset==0) *firstdevp=f;
      *lastdevp=f;
      dev_line[f]=offset++;
      mvwaddstr(w2,DEV_Y_START+dev_line[f],DEV_X_START,dev_labels[f]);
      mvwaddch(w2,DEV_Y_START+dev_line[f],DEV_X_BAR,'[');
      mvwaddch(w2,DEV_Y_START+dev_line[f],DEV_X_BAR+52,']');
    }
  }
}

//***********************************************************
int mixer()
{ int quit=0;
  int cursel;	/* currently selected device */
  int oldsel;
  int mixfd;
  int key;
  int existmask,canrecmask,isrecmask,stereomask;
  int firstdev,lastdev;
  int f;

  init(&mixfd,&existmask,&canrecmask,&isrecmask,&stereomask);

/* later things assume existmask is non-zero, so... */
  if(!existmask)
    fprintf(stderr,"mixer: /dev/mixer has no devices!\n"),exit(1);

  init_term();
  setupframe(existmask,canrecmask,&firstdev,&lastdev);
  cursel=firstdev;
  for(f=0;f<SOUND_MIXER_NRDEVICES;f++)
    if(existmask&(1<<f))
    { drawlevel(f,mixer_getlevel(mixfd,f));
//      if(canrecmask&(1<<f)) drawrec(f,isrecmask&(1<<f));
    }
  drawsel(cursel,-1);
  wrefresh(w2);

  while(!quit)
  { oldsel=cursel;
  
    key=wgetch(w3);
    switch(key)
    { case 'e': case 27:
        quit=1; 
        break;
      case 'h':
        wshow_help(w3,mixer_help);
        break;
      case 'k': case KEY_UP:
        do
          cursel--;
        while(cursel>=firstdev && !(existmask&(1<<cursel)));
        if(cursel<firstdev) cursel=lastdev;
        break;
      case 'j': case KEY_DOWN:
        do
          cursel++;
        while(cursel<=lastdev && !(existmask&(1<<cursel)));
        if(cursel>lastdev) cursel=firstdev;
        break;
      case KEY_LEFT: case '-': case '_': case '/': case '?':
        if(existmask&(1<<cursel))
        { mixer_change(mixfd,cursel,(key=='H')?-1:-2);
          drawlevel(cursel,mixer_getlevel(mixfd,cursel));
        }
        break;
      case KEY_RIGHT: case '+': case '=': case ']': case '}':
        if(existmask&(1<<cursel))
        { mixer_change(mixfd,cursel,(key=='L')?1:2);
          drawlevel(cursel,mixer_getlevel(mixfd,cursel));
        }
        break;
      case 'R'-0x40: case 'L'-0x40:	/* ^R and ^L */
        clearok(curscr,TRUE);
        break;
      default:
        break;
    }
    drawsel(cursel,oldsel);
    wrefresh(w2);
    
  }
  uninit(mixfd);
}
//This function is user to set mixer levels in non iterative mode
int cmd_mixer(int ac, char **av)
{ int mixfd;
  int existmask,canrecmask,isrecmask,stereomask;
  int f,tmp,l,r;
  int found;
  int offset_sign=0;
  char *ptr;

  init(&mixfd,&existmask,&canrecmask,&isrecmask,&stereomask);

  while(ac>1)
  { 
//lookup av[0] in dev_names[] 
    found=0;
    for(f=0;f<SOUND_MIXER_NRDEVICES;f++)
      if(existmask&(1<<f) && strcmp(av[0],dev_names[f])==0)
      { found=1;
        break;
      } 
    if(!found)
    { fprintf(stderr,"mixer: unavailable or unknown device `%s'.\n",av[0]);
      return(1);
    }
  
    ac--;
    av++;
    tmp=f;
//setting can be mono or stereo in % or offset
// see if it's an offset 
    if(av[0][0]=='+') offset_sign=1;
    if(av[0][0]=='-') offset_sign=-1;
  
//be sure to puke if level contains funny chars 
    for(f=(offset_sign?1:0);f<strlen(av[0]);f++)
      if(strchr("0123456789, \t",av[0][f])==NULL)
      { fprintf(stderr,"mixer: invalid level for device `%s'.\n",av[0]);
        exit(1);
      }
    
//now check for stereo 
    if((ptr=strchr(av[0],','))!=NULL)
    { if(offset_sign)
      { fprintf(stderr,"mixer: level offset must be mono, not stereo.\n");
        exit(1);
      }
      l=atoi(av[0]);
      r=atoi(ptr+1);
      mixer_setlevel_stereo(mixfd,tmp,l,r);
    }
    else
    {
//must be mono; default to 2 for an offset with no size specified. 
      if(offset_sign)
        mixer_change(mixfd,tmp,
          (strlen(av[0])<2)?2*offset_sign:atoi(av[0]));
      else mixer_setlevel(mixfd,tmp,atoi(av[0]));
    }
    ac--; 
    av++; 
  }
}
