/*
MP3 STATION

A plain text Linux/Unix toolset for managing mp3 playback with as little
user interaction as possible.  It was built to handle mp3 playback in a car.
Copyright (C) 2003 David Rao <louigi600@yahoo.it>

This file is part of MP3 STATION.

MP3 STATION is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MP3 STATION is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MP3 STATION; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "my_functions.h"

char PROJ_NAME='3';

//**************************************************************
int get_ug (name,ugid_arr)
char name[MAX_STR];
int ugid_arr[2];
{ struct stat stbuf;
  
  if(stat(name,&stbuf)!=0) return(-1);
  else
  { ugid_arr[0]=stbuf.st_uid;
    ugid_arr[1]=stbuf.st_gid;
  }
  return(0);
}

//**************************************************************
int set_rugid (ugid_arr)
int ugid_arr[2];
{ int ret=0;

//  if((ret=setreuid(ugid_arr[0],ugid_arr[0]))!=-1)
//    ret=setregid(ugid_arr[1],ugid_arr[1]);
  if( setgid(ugid_arr[1]) != 0 ) ret = 1;
  if( setregid(ugid_arr[1],ugid_arr[1]) != 0) ret = ret | 2;
  if( setuid(ugid_arr[0]) != 0 ) ret = ret | 8;
  if( setreuid(ugid_arr[0],ugid_arr[0]) != 0) ret = ret | 16;
  
  return(ret);
}

//**************************************************************
int drop_privs (void)
{ int euid,egid,ret;

  egid=getegid();
  euid=geteuid();
  if((ret=setreuid(euid,euid))!=-1)
    ret=setregid(egid,egid);
  return(ret);
}

//After calling this function you should make sure
//that nothing writes to stdon/out use realloc_stdioe
//**************************************************************
int be_daemon (void)
{ int pid;

  if((pid=fork())>0) exit(0);
  else if(pid==0) setsid();
  else return(pid);
}

//This closes stdin stdout stderr
//You might want to do this when running in daemon mode
//**************************************************************
void close_stdioe (void)
{ close(0);
  close(1);
  close(2);
}

//This redirects stdin stdout stderr to new file descriptors
//You might want to do this when running in daemon mode
// /dev/null and some loge file respectiveli might be a good idea
//**************************************************************
int realloc_stdioe (nfdi, nfdo, nfde)
int nfdi,nfdo,nfde;
{ int stat=0;
  int fdi,fdo,fde;
  
  if((fdi=fileno(stdin))==-1) fdi=0;
  if((fdo=fileno(stdout))==-1) fdo=1;
  if((fde=fileno(stderr))==-1) fde=2;

  if(dup2(nfdi,fdi)<0) stat=stat | 1;
  if(dup2(nfdo,fdo)<0) stat=stat | 2;
  if(dup2(nfde,fde)<0) stat=stat | 4;

  return(stat);
}

//**************************************************************
int Read_Ln (fd, buf)
int fd;
char buf[256];
{
  int i,j,k;
  char c[1];

  j=0;
  sprintf(buf,"");
  k=read(fd,c,1);
  while ( *c!='\n'&& k>0 )
  {
    if((*c!='\n') && k==1) 
    { j++;
      sprintf(buf,"%s%c",buf,*c);
    }
    k=read(fd,c,1);
  }
  return(j); 
}

//this is a better readlen but I still havetoc heckout that 
//it works fine as it counts the \n char but does not return it
int read_ln (fd, buf)
int fd;
char buf[256];
{
  int i,j,k;
  char c[1];

  j=0;
  sprintf(buf,"");
  k=read(fd,c,1);
  while ( *c!='\n'&& k>0 )
  { j++;
    if(*c!='\n') sprintf(buf,"%s%c",buf,*c);
    k=read(fd,c,1);
  }
  if((*c=='\n') && (k>0)) j++;
  return(j);
}

//see /usr/include/asm/errno.h
//**************************************************************
int Handle_Error ( err, originated_from )
int err;
char originated_from[80];
{ extern int VERBOSE;
//error n. 17 = file exists
//for now I'm regarding it as a recoverable error
  if ( err == 17 )
  {
    if (VERBOSE==1)
      printf("Warning %i while running \"%s\": %s\n",
        err,originated_from,strerror(err));
    return 0;
  }
  else
  {
    printf("Error %i while running \"%s\": %s\n",
      err,originated_from,strerror(err));
    return 1;
  }
}

//This function should look ad link destinations transparently
//as stat is being used.
//**************************************************************
int Is_Dir ( name )
char name[MAX_STR];
{ struct stat stbuf; 
  
  if(stat(name,&stbuf)!=0) return(1);
  if(S_ISDIR(stbuf.st_mode)) return(0); else return(1);
} 

//**************************************************************
int Is_File ( name )
char name[MAX_STR];
{ struct stat stbuf;

  if(stat(name,&stbuf)!=0) return(1);
  if(S_ISREG(stbuf.st_mode)) return(0); else return(1);
}

//**************************************************************
long get_newest_mtime ( name )
char name[MAX_STR];
{ struct stat stbuf;
  long newtime,ttime;

  if(stat(name,&stbuf)!=0) 
    return(-1);
  else newtime=stbuf.st_mtime;
  if(Is_Dir(name)==0)
    if((ttime=dirwalk(name,get_newest_mtime)) > newtime) newtime=ttime; 
  return(newtime);
}

//**************************************************************
char *strtolower (str,ret)
char str[MAX_STR];
char ret[MAX_STR];
{ int i;

  for(i=0;i<=strlen(str);i++)
    ret[i]=tolower(str[i]);

  return(ret);
}

//**************************************************************
char *strtoupper (str,ret)
char str[MAX_STR];
char ret[MAX_STR];
{ int i;

  for(i=0;i<=strlen(str);i++)
    ret[i]=toupper(str[i]);

  return(ret);
}

//this function walks all directory enteries 
//applying func to all enteries found
//**************************************************************
long dirwalk (dir,func)
char dir[MAX_STR];
long (*func)(char*);
{ char name[MAX_STR];
  struct dirent *dp;
  DIR *dfd;
  long ret_stat=0,t=0;
  

  if ((dfd=opendir(dir))==NULL)
  { fprintf(stderr,"Cannot open: %s\n",dir);
    return(1);
  }
  while ((dp=readdir(dfd))!=NULL)
  { if ((strcmp(dp->d_name,".")==0) || (strcmp(dp->d_name,"..")==0))
      continue;
    else if ((strlen(dir)+strlen(dp->d_name)+2)>sizeof(name))
    { fprintf(stderr,"Path name too long: %s/%s",dir,dp->d_name);
      continue;
    }
    else
    { if (strcmp(dir,".")==0) sprintf(name,"%s",dp->d_name);
      else sprintf(name,"%s/%s",dir,dp->d_name);
//      printf("%s\n",name);
      if((t=(*func)(name)) > ret_stat) ret_stat=t;
    }
  }
  return(ret_stat);
}

//see /usr/include/asm/errno.h
//**************************************************************
int msg_del_queue (queue)
char queue;
{ key_t key;
  int msgqid=-1;
  extern int VERBOSE;
  struct msqid_ds buf2;

#ifdef FPFK
  if((key=ftok((char*)FPFK,queue))!= -1)
#else
  if((key=ftok((char*)"/usr/local/bin/cm3s",queue))!= -1)
#endif
  { if(VERBOSE>1) printf("Got message key\n");
    if((msgqid=msgget(key,IPC_CREAT))!=-1) 
      if(msgctl(msgqid,IPC_RMID,&buf2)!=-1) return(0);
      else 
      { if(VERBOSE>0) printf("Error %i: %s\n",errno,perror); 
        return(-1);
      }
  }
}

//see /usr/include/asm/errno.h
//**************************************************************
int sendmsg (queue, msg)
char queue;
char msg[MAX_STR];
{ key_t key;
  int msgqid=-1;
  my_msgbuf *pmb;
  char *ps; 
  int msgsz;
  extern int VERBOSE;
  struct msqid_ds buf2;
  
  msgsz=sizeof(long)+strlen(msg)+1;
  pmb=(my_msgbuf*)malloc(msgsz);
  pmb->mtype=(long)1;
  ps=(char*)pmb + sizeof(long);
  sprintf((char*)ps,"%s",msg);
#ifdef FPFK
  if((key=ftok((char*)FPFK,queue))!= -1)
#else
  if((key=ftok((char*)"/usr/local/bin/cm3s",queue))!= -1)
#endif
  { if(VERBOSE>1) printf("Got message key\n"); 
    if((msgqid=msgget(key,IPC_CREAT))!=-1)
    { if(VERBOSE>1) printf("Queue id: %i\n",msgqid);
      if((msgsnd(msgqid,pmb,msgsz,IPC_NOWAIT|0666))!=-1)
      { if(VERBOSE>1) printf("Sent Message!\n");  
         free(pmb);
         return(0);
      }
      else if(VERBOSE>0) 
        printf("Error %i sending msg via ipc\n",errno); 
    } 
    else 
    { printf("Error %i while creating msg queue!\n",errno);
      msgctl(msgqid,IPC_STAT,&buf2);
      printf("%i,%i,%o\n",
        buf2.msg_perm.uid,
        buf2.msg_perm.gid,
        buf2.msg_perm.mode);
    }
  } else printf("Error!\n");
}

//see /usr/include/asm/errno.h
//**************************************************************
int readmsg (queue, msg)
char queue; 
char msg[MAX_STR];
{ key_t key;
  int msgqid=-1;
  int msgsz;
  struct msgbuffer {
    long mtype;     /* message type, must be > 0 */
    char mtext[MAX_STR];  /* message data */
  } msgbuf;
  extern int VERBOSE;

#ifdef FPFK
  if((key=ftok((char*)FPFK,queue))!= -1)
#else
  if((key=ftok((char*)"/usr/local/bin/cm3s",queue))!= -1)
#endif
  { if((msgqid=msgget(key,IPC_CREAT|0666))!=-1)
    { if(VERBOSE>1) printf("Queue id: %i\n",msgqid);
      if((msgsz=msgrcv(msgqid,&msgbuf,MAX_STR+4,0,IPC_NOWAIT))!=-1)
        { strncpy(msg,msgbuf.mtext,msgsz);
          return(0);
        }
      else 
      { strcpy(msg,"No message present in queue!\0");
        return(-1);
      }
    }
  }
  return(0);
}

//**************************************************************
int min(a,b)
int a,b;
{ if(a<b) return(a);
  else return(b);
}

//**************************************************************
int max(a,b)
int a,b;
{ if(a<b) return(b);
  else return(a);
}

//**************************************************************
char *path2tok(src,dst)
char src[MAX_STR],dst[MAX_STR];
{ long i,d;
  struct stat stbuf;

  stat(src,&stbuf);
  i=stbuf.st_ino;
  d=stbuf.st_dev;
  sprintf(dst,"%x-%x",d,i);
  
  return(dst);
}

//This function cuts from a src string any unnecessary blancs or tabs
//and returns dst
//**************************************************************
char *cut_space(dst, src)
char dst[MAX_STR],src[MAX_STR];
{ int i=0,j=0;
  int prev=-1;

  for(;i<=strlen(src);i++)
  { switch(src[i])
    { case '\f': case '\n': case '\r': case '\t': case '\v': 
        if(src[i + 1] > 32) { dst[j]=':'; ++j; }
        prev=' ';
        continue;
      break;
      case ' ':
        if((src[i + 1]==' ')||(src[i + 1]<14)) { prev=' '; continue; }
        else if(prev==' ') { dst[j]=':'; ++j; }
      break;
      default:
        prev=src[i];
      break;
    }
    dst[j]=src[i];
    ++j;
  }
  return(dst);
}

//remove blanks and tabs from srd and return dst
char *rm_space(dst, src)
char dst[MAX_STR],src[MAX_STR];
{ int i=0,j=0;

  for(;i<=strlen(src);i++)
  { switch(src[i])
    { case '\f': case '\n': case '\r': case '\t': case '\v': case ' ': 
        continue;
      break;
      default:
        dst[j]=src[i];
        ++j;
      break;
    }
  }
  return(dst);
}

//This function returnes any the last pid of any processe who have the
//given name put are not currnet proecess nor children of currnet proecess
//**************************************************************
int other_running(name)
char name[256];
{ FILE *p;
  char buf[256],cmd[256];
  int mpid,opid=0;

  mpid=getpid();

  sprintf(cmd,
"ps -eo pid,ppid,command | grep \"%s\" | grep -vE \"grep|vi |less |^ *%i | %i\"",
    name,mpid);
  if( p=popen(cmd,"r"))
    while( fgets(buf, 255, p) != NULL ) { opid=atoi(buf); }

  return(opid);
}

//This funtion does something like fuser 
//it returns > 0 if there are pids using pathname
//**************************************************************
int fuser(pathname)
char pathname[MAX_STR];
{ FILE *p;
  char buf[256],cmd[256];
  int upid=0;
  sprintf(cmd,"fuser %s",pathname);
  
  if( p=popen(cmd,"r"))
    while( fgets(buf, 255, p) != NULL ) upid++;

  return(upid);
}

//this is an alternative sleep that takes as input 1/64 second time units
//it will try to use in cascade nansleep, RTC, normal sleep
//if it fails to open /dev/rtc (for eg vmware locks the device)
//it will round up the time given to the nearest second and do a normal
//sleep with the rounded time
int sleep64(counter)
int counter;
{ int i, fd, retval, irqcount = 0;
  unsigned long tmp=64, data;
#ifdef NANOSLEEP
//1/64 of a second expressed in nanoseconds
  long utn=15625000; 
  struct timespec tspec;
  
  tspec.tv_nsec= (utn * counter) - 1;
  if(nanosleep(&tspec,NULL)==0) return(0); 
#endif

#ifdef RTC
  fd = open ("/dev/rtc", O_RDONLY);
  if(fd == -1)
  { counter=(counter + 64) / 64;
    sleep(counter);
    return(1);
  }

  retval = ioctl(fd, RTC_IRQP_SET, tmp);
  if(retval == -1)
  { counter=(counter + 64) / 64;
    sleep(counter);
    close(fd);
    return(1);
  }
  retval = ioctl(fd, RTC_PIE_ON, 0);
  if(retval == -1)
  { counter=(counter + 64) / 64;
    sleep(counter);
    close(fd);
    return(1);
  }
  for(i=0; i<counter;  i++)
  { retval = read(fd, &data, sizeof(unsigned long));
    irqcount++;
  }
  retval = ioctl(fd, RTC_PIE_OFF, 0);
  close(fd);
  return(0);
#else
  counter=(counter + 64) / 64;
  sleep(counter);
   return(1);
#endif
}

int atox(char *hex_str)
{ char tmp_str[20];
  int i,j,k=1;
  int retval=0;
  int c;

  if(strncmp(hex_str,"0x",2) != 0) return(-1);
  for(i=strlen(strcpy(tmp_str,&hex_str[2]))-1;i>=0;i--)
  { if(strncmp(&tmp_str[i],"a",1)==0) j=10;
    else if(strncmp(&tmp_str[i],"b",1)==0) j=11;
    else if(strncmp(&tmp_str[i],"c",1)==0) j=12;
    else if(strncmp(&tmp_str[i],"d",1)==0) j=13;
    else if(strncmp(&tmp_str[i],"e",1)==0) j=14;
    else if(strncmp(&tmp_str[i],"f",1)==0) j=15;
    else j=atoi(&tmp_str[i]);
    tmp_str[i]='\0';
    retval=retval + j * k;
    k=k * 16;
  }
  return(retval);
}

//it appearsthat whencpu has nothing to do it makes mpg123
//produce a low freq buz on the power lines that get into my amplifier
//this function just keeps the cpu busy w/o doing anything usefull
void eat_cpu()
{ long i=123,j=321,k;

  while(1) k= i * j;
}
