/*
MP3 STATION

A plain text Linux/Unix toolset for managing mp3 playback with as little
user interaction as possible.  It was built to handle mp3 playback in a car.
Copyright (C) 2003 David Rao <louigi600@yahoo.it>

This file is part of MP3 STATION.

MP3 STATION is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

MP3 STATION is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MP3 STATION; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <libgen.h>
#include <stdlib.h>

//Typedefs
//***********************************************************************
//pointer to next element in option list
typedef struct opt_node *opt_ptr;

//parsed option list structure
typedef struct opt_node {
  int short_name;
  char args[256];
  opt_ptr next;
} my_opt_ind;


//External variables
//***********************************************************************
extern char pt[]; 	//progtype server or client
extern char MP3_DIR[];
extern char TMP_DIR[];
extern char STAT_PIPE[];
extern char CMD_PIPE[];
extern char SETTINGS[];
extern char PLAYLIST[];
extern char LOG[];
extern char TPL[];
extern char OFFSET[];
extern char VERSION[];
extern int BEDAEMON;
extern int RANDOM;
extern int BACKPLAY;
extern int REPEAT;
extern long START;
extern int NEXT;
extern int BUFFER;
extern int VERBOSE;

//Global Variables
//***********************************************************************
//Since different platforms have such different getopt implementations it 
//became necessary to write my own portable implementation
//my_option structrute has the folowing definition and it's fealds are as such:
//long_name is the long option name witout the leading -- that is mandatory 
//at comand prompt
//has_arg can be 0,1,2 0=no param expected, 1=mandatory param 2=optional param
//N.B. for the moment optional parameters will not accept files or directories
//if the optional argument is the last argument in argv 
//short_name is the one character without the leading - representing the short option
struct my_option {
  const char *long_name;
  int has_arg;
  int short_name;
};

struct my_option long_options[] = {
  {"back",0,'B'},
  {"buffer",1,'b'},
  {"mp3dir",1,'d'},
  {"help",0,'h'},
  {"daemon",0,'D'},
  {"log",1,'l'},
  {"next",0,'n'},
  {"repeat", 0,'R'},
  {"random", 2,'r'},
  {"start", 1,'s'},
  {"tmp",1,'t'},
  {"version", 0,'V'},
  {"verbose", 1,'v'},
  {0, 0, 0} };

//static char help [][80] = {
char option_help[][80]={
  " <options> <playlist>",
  "-B, --back             : Proceed the playlist backward.",
  "-b, --buffer <kb size> : Activare buffer on audio output stream.",
  "-d, --mp3dir    <path> : Directory where to look for mp3 and playlists.",
  "-D, --daemon           : Rus as daemon.",
  "-h, --help             : Show this help message.",
  "-l, --log   <log_file> : When running as daemon log to an alternate file.",
  "-n, --next             : Start from last track + 1.",
  "-R, --repeat           : Repeat playinf playlist untill new list il loaded.",
  "-r, --random    [seed] : Play list in random mode [seed random generator].",
  "-s, --start  <trackno> : Start playind playlist form trackno.",
  "-t, --tmpdir    <path> : Directory where to write temporary files.",
  "-V, --version          : Show version and exit.",
  "-v, --verbose  <level> : Higher level produces more output.",
  "N.B. Be carefull if you are seeding -r jointly with -s option.",
  "When using a buffer it does not make sense to use one smaller than 1/2 Mb.",
  "When running as daemon default log is in tmpdir/log.",
  "It is good practise to put -D option first, this insures no output to term",
  ""
};

char pn[1024];
opt_ptr head;

//Function declarations
//***********************************************************************
void Show_help();
opt_ptr oialloc();
opt_ptr parse_options (int ac, char **av);
opt_ptr order_opt_ind(opt_ptr start);
void get_options (int ac, char **av);
void free_oilist(opt_ptr start);


//Function definitions
//***********************************************************************
void Show_help ()
{
  int i;
  
  printf("%s %s\n",pn,option_help[0]);
  for (i=1;i<(sizeof(option_help)/sizeof(option_help[0]));i++)
    printf("%s\n",option_help[i]);
  exit(0);
}

//***********************************************************************
//allocate memory for my_opt_ind 
opt_ptr oialloc()
{ return ((opt_ptr) malloc(sizeof(my_opt_ind)));
}

//***********************************************************************
//free my_opt_ind list
void free_oilist(start)
opt_ptr start;
{ opt_ptr node,tnode;

  node=start;
  while(node->next != NULL)
  { tnode=node;
    node=node->next;
    free(tnode);
  } 
  free(node);
}

//***********************************************************************
//move any parameters at the end of option ind list
opt_ptr order_opt_ind(start)
opt_ptr start;
{ opt_ptr nlast,last,next,prev,node;

  node=start;
  do
    last=node;
  while((node=node->next) != NULL);
  
  node=start;
  prev=start;
  nlast=last;
  while(node!=last)
  { next=node->next;
    if(node->short_name==0)
    { prev=node->next;
      if(node=start) start=node->next;
      node->next=NULL;
      nlast->next=node;
      nlast=node;
    }
    else prev=node;
    node=next;
  }
  return(start); 
}

//***********************************************************************
//parse options ans parameters, check arguments, and show help msg if 
//something is wrong
opt_ptr parse_options (int ac, char **av)
{ int i,j,hit,c;
  char lopt[80],sopt[3];
  opt_ptr last,node;

  head=(opt_ptr)oialloc();
  head->next=NULL;
  node=head;
  for(i=1;i<ac;i++)
  { hit=0;
    for(j=0;j<(sizeof(long_options)/sizeof(long_options[0]))-1;j++)
    { sprintf(sopt,"-%c",long_options[j].short_name);
      sprintf(lopt,"--%s",long_options[j].long_name);
//If it is a long option
      if(strcmp(lopt,av[i])==0) 
      { hit=1;
        node->short_name=long_options[j].short_name;
      }
//Else if it ia s short option
      else if(strcmp(sopt,av[i])==0) 
      { hit=1;
        node->short_name=long_options[j].short_name;
      }

      if(hit==1) 
      { switch (long_options[j].has_arg)
        { case 0:
            strcpy(node->args,"");
            break;
          case 1:
            if(ac==i + 1) 
            { printf("Missing required parameter for option %s !!!\n",av[i]);
              Show_help(); 
            } 
            else if(strncmp(av[i + 1],"-",1)==0)
            { printf("Missing required parameter for option %s !!!\n",av[i]);
              Show_help();
            }
            else strcpy(node->args,av[i + 1]);
            ++i;
            break;
          case 2:
//It it is an optional parameter then it will acceop it only if it is not
//the last one avalible
            if(ac>i + 2) 
            { if(strncmp(av[i + 1],"-",1)!=0)
              { strcpy(node->args,av[i + 1]);
                ++i;
              }
              else strcpy(node->args,"");
            }
            else if(ac>i +1) 
            { if((Is_File(av[i + 1])!=0) || (Is_Dir(av[i + 1])!=0) || 
                 (strncmp(av[i + 1],"-",1)!=0)) 
              { strcpy(node->args,av[i + 1]);
                ++i;
              }
              else strcpy(node->args,"");
            }
            else strcpy(node->args,"");
            break;
        }
        last=node;
        if(ac>i + 1)
        { node=oialloc();
          node->next=NULL;
          last->next=node;
        }
//        printf("-%c %s:%i %i\n",last->short_name,last->args,last,last->next);
        break;
      }
    }
//this is an unrecognised option
    if((hit==0) && (strncmp(av[i],"-",1)==0)) 
    { printf("%s: unrecognised option !!!\n",av[i]);
      Show_help();
    }
    else if(hit==0)
    { node->short_name=0;
      strcpy(node->args,av[i]);
      last=node;
      if(ac>i + 1)
      { node=oialloc();
        node->next=NULL;
        last->next=node;
      }
    }
  }
  head=order_opt_ind(head);
  return(head);
}

//***********************************************************************
//looking at the options and setting necessary variables
void get_options (int ac, char **av)
{
  int c,i;
  int digit_optind = 0;
  char optistr[80],tmpstr[80];
  opt_ptr node;

  strcpy(pn,basename(av[0]));

  head=parse_options(ac,av);
  node=head;
  do
  { switch(node->short_name)
    { case 'B':
        BACKPLAY=1;
      break;

      case 'b':
        BUFFER=atoi(node->args);
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Activating %i Kb audio buffer.\n",BUFFER);
      break;

      case 'd':
        strcpy(MP3_DIR,node->args);
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Starting up with %s as default mp3 dir.\n",MP3_DIR);
      break;

      case 'D':
        BEDAEMON=1;
        if(VERBOSE>1) printf("Starting as daemon!\n");
      break;

      case 'h':
        Show_help();
      break;
 
      case 'l':
        strcpy(LOG,node->args);
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Will be logging to: %s instead of default.\n",LOG);
      break;

      case 'n':
        NEXT=1;
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Starting from last track + 1.\n");
      break;

      case 'R':
        REPEAT=1;
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Will play list in repeat mode.\n");
      break;
      
      case 'r':
        if(strlen(node->args)>0) RANDOM=atoi(node->args);
        else RANDOM=1;
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Seeding random generator with %i.\n",RANDOM);
      break;

      case 's':
        START=atol(node->args);
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Starting to play from track %i.\n",START);
      break;

      case 't':
        strcpy(TMP_DIR,node->args);
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Using %s to save settings and temp files.\n",TMP_DIR);
      break;

      case 'V':
        printf("%s\n",VERSION);
        exit(0);
      break;

      case 'v':
        VERBOSE=atoi(node->args);
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Using verbose level %i.\n",VERBOSE);
      break;

      case '?':
        Show_help();
      break;

      case 0:
        strcpy(PLAYLIST,node->args);
        if((VERBOSE>0) && (BEDAEMON==0))
          printf("Loading playlist %s.\n",PLAYLIST);
        if(node->next!=NULL) Show_help();
      break;
    }
  } while((node=node->next) != NULL);
  free_oilist(head);
}
